@@ -64,6 +64,8 @@ typedef struct
 	bool		double_header;
 	char		header_mode;
 	bool		ignore_short_rows;
+	int			trim_width;
+	int			trim_rows;
 } PrintConfigType;
 
 static void pb_putc_repeat(PrintbufType *printbuf, int n, int c);
@@ -414,16 +416,17 @@ is_header(RowBucketType *rb)
 static char *
 pb_put_line(char *str, bool multiline, PrintbufType *printbuf)
 {
-	char   *nextline = NULL;
+	char	   *nextline = NULL;
 
 	if (multiline)
 	{
-		char   *ptr = str;
-		int		size = 0;
-		int		chrl;
+		char	   *ptr = str;
+		int			size = 0;
 
 		while (*ptr)
 		{
+			int			chrl;
+
 			if (*ptr == '\n')
 			{
 				nextline = ptr + 1;
@@ -443,6 +446,76 @@ pb_put_line(char *str, bool multiline, PrintbufType *printbuf)
 	return nextline;
 }
 
+static char *
+pb_put_line_trim_width(char *str, bool multiline, PrintbufType *printbuf, int width)
+{
+	char	   *ptr = str;
+	char	   *nextline = NULL;
+	int			str_width = 0;
+	int			str_size = 0;
+	int			charwidth;
+	int			charsize;
+
+	while (*ptr)
+	{
+		if (*ptr == '\n')
+		{
+			nextline = ptr + 1;
+			break;
+		}
+		else if (*ptr == '\t')
+		{
+			int aux_str_width = str_width;
+
+			do
+			{
+				aux_str_width++;
+			} while (aux_str_width % 8 != 0);
+
+			charsize = 1;
+			charwidth = aux_str_width - str_width;
+		}
+		else if (use_utf8)
+		{
+			charsize = charlen(ptr);
+			charwidth = utf_dsplen(ptr);
+		}
+		else
+		{
+			charsize = 1;
+			charwidth = 1;
+		}
+
+		if (str_width + charwidth <= width)
+		{
+			ptr += charsize;
+			str_size += charsize;
+			str_width += charwidth;
+		}
+		else
+			break;
+	}
+
+	pb_write(printbuf, str, str_size);
+
+	if (use_utf8)
+		pb_write(printbuf, "\342\200\245", 3); /* ‥ */
+
+	while (str_width++ < width)
+		pb_write(printbuf, " ", 1);
+
+	if (multiline)
+	{
+		while (*ptr && *ptr != '\n')
+			ptr += use_utf8 ? charlen(str) : 1;
+
+		if (*ptr == '\n')
+			nextline = ptr + 1;
+	}
+
+	return nextline;
+}
+
 /*
  * Print formatted data loaded inside RowBuckets
  */
@@ -453,12 +526,12 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 				   PrintDataDesc *pdesc,
 				   char *title)
 {
-	bool	is_last_column_multiline = pdesc->multilines[pdesc->nfields - 1];
-	int		last_column_num = pdesc->nfields - 1;
-	int		printed_rows = 0;
-	char	linestyle = pconfig->linestyle;
-	int		border = pconfig->border;
-	char	buffer[20];
+	bool		is_last_column_multiline = pdesc->multilines[pdesc->nfields - 1];
+	int			last_column_num = pdesc->nfields - 1;
+	int			printed_rows = 0;
+	char		linestyle = pconfig->linestyle;
+	int			border = pconfig->border;
+	char		buffer[20];
 
 	printbuf->printed_headline = false;
 	printbuf->flushed_rows = 0;
@@ -474,15 +547,14 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 
 	while (rb)
 	{
-		int		i;
+		int			i;
 
 		for (i = 0; i < rb->nrows; i++)
 		{
-			int		j;
-			bool	isheader = false;
+			int			j;
 			RowType	   *row;
-			bool	more_lines = true;
-			bool	multiline = rb->multilines[i];
+			bool		more_lines = true;
+			bool		multiline = rb->multilines[i];
 			char	   *fields[1024];
 			int			multiline_lineno;
 
@@ -490,11 +562,13 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 			if (pconfig->ignore_short_rows && rb->rows[i]->nfields != pdesc->nfields_all)
 				continue;
 
-			multiline_lineno = 0;
+			multiline_lineno = 1;
 			row = rb->rows[i];
 
 			while (more_lines)
 			{
+				bool		isheader = false;
+
 				more_lines = false;
 
 				if (border == 2)
@@ -511,10 +585,8 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 
 				for (j = 0; j < pdesc->nfields; j++)
 				{
-					int		width;
-					int		spaces;
-					char   *field;
-					bool	_more_lines = false;
+					char	   *field;
+					bool		_more_lines = false;
 
 					if (j > 0)
 					{
@@ -529,7 +601,7 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 
 					if (pdesc->columns_map[j] < row->nfields)
 					{
-						if (multiline_lineno == 0)
+						if (multiline_lineno == 1)
 						{
 							field = row->fields[pdesc->columns_map[j]];
 							fields[j] = NULL;
@@ -542,6 +614,7 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 
 					if (field && *field != '\0')
 					{
+						int			width;
 						bool	left_align = pdesc->types[j] != 'd';
 
 						if (!use_utf8)
@@ -554,8 +627,7 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 							{
 								if (*ptr == '\n')
 								{
-									more_lines = true;
-									ptr += 1;
+									_more_lines = true;
 									break;
 								}
 								else if (*ptr == '\t')
@@ -578,40 +650,57 @@ pb_print_rowbuckets(PrintbufType *printbuf,
 						else
 						{
 							if (multiline)
-							{
-								width = utf_string_dsplen_multiline(field, INT_MAX, &_more_lines, true, NULL, NULL);
-								more_lines |= _more_lines;
-							}
+								width = utf_string_dsplen_multiline(field, INT_MAX, &_more_lines, true, NULL, NULL, 0);
 							else
 								width = utf_string_dsplen(field, INT_MAX);
 						}
 
-						spaces = pdesc->widths[j] - width;
-
-						/*
-						 * The display width can be canculated badly when labels or
-						 * displayed string has some special or invisible chars. Here
-						 * is simple ugly fix - the number of spaces cannot be negative.
-						 */
-						if (spaces < 0)
-							spaces = 0;
-
-						/* left spaces */
-						if (isheader)
-							pb_putc_repeat(printbuf, spaces / 2, ' ');
-						else if (!left_align)
-							pb_putc_repeat(printbuf, spaces, ' ');
-
 						if (multiline)
-							fields[j] = pb_put_line(field, multiline, printbuf);
-						else
-							(void) pb_put_line(field, multiline, printbuf);
+						{
+							if (pconfig->trim_rows > 0 && multiline_lineno == pconfig->trim_rows)
+								_more_lines = false;
+							else
+								more_lines |= _more_lines;
+						}
 
-						/* right spaces */
-						if (isheader)
-							pb_putc_repeat(printbuf, spaces - (spaces / 2), ' ');
-						else if (left_align)
-							pb_putc_repeat(printbuf, spaces, ' ');
+						if (pconfig->trim_width > 0 && pconfig->trim_width < width)
+						{
+							if (multiline)
+								fields[j] = pb_put_line_trim_width(field, multiline, printbuf, pconfig->trim_width);
+							else
+								(void) pb_put_line_trim_width(field, multiline, printbuf, pconfig->trim_width);
+						}
+						else
+						{
+							int			spaces;
+
+							spaces = pdesc->widths[j] - width;
+
+							/*
+							 * The display width can be canculated badly when labels or
+							 * displayed string has some special or invisible chars. Here
+							 * is simple ugly fix - the number of spaces cannot be negative.
+							 */
+							if (spaces < 0)
+								spaces = 0;
+
+							/* left spaces */
+							if (isheader)
+								pb_putc_repeat(printbuf, spaces / 2, ' ');
+							else if (!left_align)
+								pb_putc_repeat(printbuf, spaces, ' ');
+
+							if (multiline)
+								fields[j] = pb_put_line(field, multiline, printbuf);
+							else
+								(void) pb_put_line(field, multiline, printbuf);
+
+							/* right spaces */
+							if (isheader)
+								pb_putc_repeat(printbuf, spaces - (spaces / 2), ' ');
+							else if (left_align)
+								pb_putc_repeat(printbuf, spaces, ' ');
+						}
 					}
 					else
 						pb_putc_repeat(printbuf, pdesc->widths[j], ' ');
@@ -773,7 +862,9 @@ postprocess_fields(int nfields,
 				   LinebufType *linebuf,
 				   bool ignore_short_rows,
 				   bool reduced_sizes,			/* the doesn't calculate ending zero */
-				   bool *is_multiline_row)
+				   bool *is_multiline_row,
+				   int trim_width,
+				   int trim_rows)
 {
 	bool		malformed;
 	size_t		width;
@@ -840,7 +931,11 @@ postprocess_fields(int nfields,
 												&multiline,
 												false,
 												&digits,
-												&total);
+												&total,
+												trim_rows);
+
+		if (trim_width > 0 && width > trim_width)
+			width = use_utf8 ? trim_width + 1 : trim_width;
 
 		/* skip first possible header row */
 		if (linebuf->processed > 0)
@@ -1103,13 +1198,13 @@ read_tsv(RowBucketType *rb,
 		}
 		else
 		{
-			int		i;
 
 			if (linebuf->used > 0)
 			{
-				char   *locbuf;
+				char	   *locbuf;
 				RowType	   *row;
-				bool	multiline = false;
+				bool		multiline = false;
+				int			i;
 
 				append_char(linebuf, '\0');
 				linebuf->sizes[nfields++] = size + 1;
@@ -1136,7 +1231,9 @@ read_tsv(RowBucketType *rb,
 								   linebuf,
 								   ignore_short_rows,
 								   false,
-								   &multiline);
+								   &multiline,
+								   opts->csv_trim_width,
+								   opts->csv_trim_rows);
 
 				rb->multilines[rb->nrows] = multiline;
 				rb->rows[rb->nrows++] = row;
@@ -1391,7 +1488,9 @@ read_csv(RowBucketType *rb,
 							   linebuf,
 							   ignore_short_rows,
 							   true,
-							   &multiline);
+							   &multiline,
+							   opts->csv_trim_width,
+							   opts->csv_trim_rows);
 
 			rb->multilines[rb->nrows] = multiline;
 			rb->rows[rb->nrows++] = row;
@@ -1479,7 +1578,6 @@ read_and_format(Options *opts, DataDesc *desc, StateData *state)
 	lb_free(desc);
 	memset(desc, 0, sizeof(DataDesc));
 
-	name = (char *) get_input_file_basename();
 	if ((name = (char *) get_input_file_basename()))
 	{
 		strncpy(desc->filename, name, 64);
@@ -1529,6 +1627,9 @@ read_and_format(Options *opts, DataDesc *desc, StateData *state)
 	pconfig.header_mode = opts->csv_header;
 	pconfig.ignore_short_rows = opts->ignore_short_rows;
 
+	pconfig.trim_width = opts->csv_trim_width;
+	pconfig.trim_rows = opts->csv_trim_rows;
+
 	memset(&rowbuckets, 0, sizeof(RowBucketType));
 
 	rowbuckets.allocated = false;
@@ -1536,7 +1637,10 @@ read_and_format(Options *opts, DataDesc *desc, StateData *state)
 	rowbuckets.next_bucket = NULL;
 
 	if (opts->querystream && !query)
+	{
+		free(linebuf.buffer);
 		return false;
+	}
 
 	if (query)
 	{
